@@ -3,7 +3,7 @@
  *
  * PPTP Control Message packet reading, formatting and writing.
  *
- * $Id: ctrlpacket.c,v 1.42 1999/08/07 21:29:26 luyer Exp $
+ * $Id: ctrlpacket.c,v 1.3 1999/12/23 21:43:33 patl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -423,12 +423,12 @@ void deal_out_call(unsigned char *packet, unsigned char *rply_packet, ssize_t *
 	 * still performs at over 115200, eg, 60kbyte/sec and higher observed.
 	 */
 	out_call_rply.speed = out_call_rqst->max_bps;
-	/* lets match their window size for now... was htons(PCKT_RECV_WINDOW_SIZE)
-	 */
-	out_call_rply.pckt_recv_size = out_call_rqst->pckt_recv_size;
+	out_call_rply.pckt_recv_size = hton16(PCKT_RECV_WINDOW_SIZE);
 	if(pptpctrl_debug)
-		syslog(LOG_DEBUG, "CTRL: Set parameters to %d maxbps, %d window size",
-			ntohs(out_call_rply.speed), ntohs(out_call_rply.pckt_recv_size));
+		syslog(LOG_DEBUG, "CTRL: %d min_bps, %d max_bps, %d window size",
+                       ntohs(out_call_rqst->min_bps),
+                       ntohs(out_call_rqst->max_bps),
+                       ntohs(out_call_rply.pckt_recv_size));
 	out_call_rply.pckt_delay = htons(PCKT_PROCESS_DELAY);
 	out_call_rply.channel_id = htonl(CHANNEL_ID);
 	COPY_CTRL_PACKET(out_call_rply, rply_packet, rply_size);
