@@ -4,7 +4,7 @@
  * This file contains some tuneable parameters, most of which can be overriden
  * at run-time (note, MAX_CONNECTIONS can't!).
  *
- * $Id: defaults.h,v 1.28 1999/09/09 07:32:40 luyer Exp $
+ * $Id: defaults.h,v 1.2 1999/12/17 18:57:30 patl Exp $
  */
 
 #ifndef _PPTPD_DEFAULTS_H
@@ -59,4 +59,7 @@
 #define LISTEN_KEYWORD			"listen"
 #define PIDFILE_KEYWORD			"pidfile"
 
+/* Logging facility */
+#define LOG_PPTP                        LOG_LOCAL2
+
 #endif	/* !_PPTPD_DEFAULTS_H */
