@@ -3,7 +3,7 @@
  *
  * PPTP control connection between PAC-PNS pair
  *
- * $Id: pptpctrl.c,v 1.75 1999/09/09 07:32:41 luyer Exp $
+ * $Id: pptpctrl.c,v 1.4 1999/12/29 16:06:41 patl Exp $
  */
 
 #ifdef HAVE_CONFIG_H
@@ -117,7 +117,7 @@ int main(int argc, char **argv)
 	char *pppaddrs[3] = { pppLocal, pppRemote, ipx };
 
 	/* open a connection to the syslog daemon */
-	openlog("pptpd", LOG_PID, LOG_DAEMON);
+	openlog("pptpd", LOG_PID, LOG_PPTP);
 
 	/* autoreap if supported */
 	signal(SIGCHLD, SIG_IGN);
@@ -209,203 +209,233 @@ int main(int argc, char **argv)
  */
 static void pptp_handle_ctrl_connection(char **pppaddrs, struct in_addr *inetaddrs)
 {
+  /* For echo requests used to check link is alive */
+  int echo_wait = FALSE;		/* Waiting for echo? */
+  u_int32_t echo_count = 0;	/* Sequence # of echo */
+  time_t echo_time;		/* Time last echo req sent */
+  struct timeval idleTime;	/* How long to select() */
 
-	/* For echo requests used to check link is alive */
-	int echo_wait = FALSE;		/* Waiting for echo? */
-	u_int32_t echo_count = 0;	/* Sequence # of echo */
-	time_t echo_time;		/* Time last echo req sent */
-	struct timeval idleTime;	/* How long to select() */
-
-	/* General local variables */
-	ssize_t rply_size;		/* Reply packet size */
-	fd_set fds;			/* For select() */
-	int maxfd = clientSocket;	/* For select() */
-	int send_packet;		/* Send a packet this time? */
+  /* General local variables */
+  ssize_t rply_size;		/* Reply packet size */
+  fd_set readfds;		/* For select() */
+  fd_set writefds;              /* For select() */
+  int maxfd = clientSocket;	/* For select() */
+  int send_packet;		/* Send a packet this time? */
 #if BSDUSER_PPP || SLIRP
 /* not needed by stuff which uses socketpair() in startCall() */
 #define init 1
 #else
-	int init = 0;			/* Has pppd initialized the pty? */
+  int init = 0;			/* Has pppd initialized the pty? */
 #endif
-	int pty_fd = -1;		/* File descriptor of pty */
-	int gre_fd = -1;		/* Network file descriptor */
+  struct gre_state *gre = NULL;
 
-	unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
-	unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
+  unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
+  unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
+  u_int16_t max_window_size;
+  int new_pty_fd = -1;
 
-	FD_ZERO(&fds);
+  for (;;) {
+    FD_ZERO(&readfds);
+    FD_ZERO(&writefds);
 
-	for (;;) {
+    /* always willing to read from the control channel */
+    FD_SET (clientSocket, &readfds);
 
-		FD_SET(clientSocket, &fds);
-		if (pty_fd != -1)
-			FD_SET(pty_fd, &fds);
-		if (gre_fd != -1 && init)
-			FD_SET(gre_fd, &fds);
+    /* always willing to read from the GRE socket */
+    if (gre && gre->gre_fd != -1 && init)
+      FD_SET (gre->gre_fd, &readfds);
 
-		/* set timeout */
-		if (encaps_gre(-1, NULL, 0) || decaps_hdlc(-1, NULL, 0))
-			idleTime.tv_sec = 0;
-		else
-			idleTime.tv_sec = IDLE_WAIT;
-		idleTime.tv_usec = 0;
+    /* Willing to read from the PTY if the buffer is empty */
+    if (gre && gre->pty_fd != -1
+        && gre->pty_bytes_processed == gre->pty_bytes_read)
+      FD_SET (gre->pty_fd, &readfds);
 
-		/* default: do nothing */
-		send_packet = FALSE;
+    /* want to write to the GRE socket if we have a packet ready */
+    if (gre && gre->gre_fd != -1 && gre->gre_packet_ready)
+      FD_SET (gre->gre_fd, &writefds);
 
-		switch (select(maxfd + 1, &fds, NULL, NULL, &idleTime)) {
-		case -1:	/* Error with select() */
-			if (errno != EINTR)
-				syslog(LOG_ERR, "CTRL: Error with select(), quitting");
-			goto leave_clear_call;
+    /* want to write to the PTY if the output buffer is nonempty */
+    if (gre && gre->pty_fd != -1 && init
+        && gre->pty_output_size != gre->pty_bytes_written)
+      FD_SET (gre->pty_fd, &writefds);
 
-		case 0:
-			if (decaps_hdlc(-1, NULL, 0)) {
-				if(decaps_hdlc(-1, encaps_gre, gre_fd))
-					syslog(LOG_ERR, "CTRL: GRE re-xmit failed");
-			} else if (encaps_gre(-1, NULL, 0))
-				/* Pending ack and nothing else to do */
-				encaps_gre(gre_fd, NULL, 0);	/* send ack with no payload */
-			else if (echo_wait != TRUE) {
-				/* Timeout. Start idle link detection. */
-				echo_count++;
-				if (pptpctrl_debug)
-					syslog(LOG_DEBUG, "CTRL: Sending ECHO REQ id %d", echo_count);
-				time(&echo_time);
-				make_echo_req_packet(rply_packet, &rply_size, echo_count);
-				echo_wait = TRUE;
-				send_packet = TRUE;
-			}
-			break;
+    /* set timeout */
+    /* FIXME: need to do adaptive timeouts / windowing */
+    idleTime.tv_sec = IDLE_WAIT;
+    idleTime.tv_usec = 0;
 
-		default:
-			break;
-		}
+    /* default: do nothing */
+    send_packet = FALSE;
 
-		/* detect startup of pppd */
+    switch (select(maxfd + 1, &readfds, &writefds, NULL, &idleTime)) {
+    case -1:	/* Error with select() */
+      if (errno != EINTR)
+        syslog(LOG_ERR, "CTRL: Error with select(), quitting");
+      goto leave_clear_call;
+
+    case 0:
+      if (echo_wait != TRUE) {
+        /* Timeout. Start idle link detection. */
+        echo_count++;
+        if (pptpctrl_debug)
+          syslog(LOG_DEBUG, "CTRL: Sending ECHO REQ id %d", echo_count);
+        time(&echo_time);
+        make_echo_req_packet(rply_packet, &rply_size, echo_count);
+        echo_wait = TRUE;
+        send_packet = TRUE;
+      }
+      break;
+
+    default:
+      break;
+    }
+
+    /* detect startup of pppd */
 #ifndef init
-		if (!init && pty_fd != -1 && FD_ISSET(pty_fd, &fds))
-			init = 1;
+    if (!init && gre && gre->pty_fd != -1 && FD_ISSET(gre->pty_fd, &readfds))
+      init = 1;
 #endif
 
-		/* handle actual packets */
+    /* handle actual packets */
 
-		/* send from pty off via GRE */
-		if (pty_fd != -1 && FD_ISSET(pty_fd, &fds) && decaps_hdlc(pty_fd, encaps_gre, gre_fd) < 0) {
-			syslog(LOG_ERR, "CTRL: PTY read or GRE write failed (pty,gre)=(%d,%d)", pty_fd, gre_fd);
-			break;
-		}
-		/* send from GRE off to pty */
-		if (gre_fd != -1 && FD_ISSET(gre_fd, &fds) && decaps_gre(gre_fd, encaps_hdlc, pty_fd) < 0) {
-			syslog(LOG_ERR, "CTRL: GRE read or PTY write failed (gre,pty)=(%d,%d)", gre_fd, pty_fd);
-			break;
-		}
-		/* handle control messages */
+    /* send from pty off via GRE */
+    if (gre) {
+      if (do_gre_to_pty (gre) < 0) {
+        syslog(LOG_ERR,
+               "CTRL: PTY read or GRE write failed (pty,gre)=(%d,%d)",
+               gre->pty_fd, gre->gre_fd);
+        break;
+      }
 
-		if (FD_ISSET(clientSocket, &fds)) {
-			send_packet = TRUE;
-			switch (read_pptp_packet(clientSocket, packet, rply_packet, &rply_size)) {
-			case 0:
-				syslog(LOG_ERR, "CTRL: CTRL read failed");
-				goto leave_drop_call;
+      if (do_pty_to_gre (gre) < 0) {
+        syslog(LOG_ERR,
+               "CTRL: GRE read or PTY write failed (gre,pty)=(%d,%d)",
+               gre->gre_fd, gre->pty_fd);
+        break;
+      }
+    }
 
-			case -1:
-				send_packet = FALSE;
-				break;
+    if (FD_ISSET(clientSocket, &readfds)) {
+      send_packet = TRUE;
+      switch (read_pptp_packet(clientSocket, packet,
+                               rply_packet, &rply_size)) {
+      case 0:
+        syslog(LOG_ERR, "CTRL: CTRL read failed");
+        goto leave_drop_call;
 
-			case STOP_CTRL_CONN_RQST:
-				if (pptpctrl_debug)
-					syslog(LOG_DEBUG, "CTRL: Received STOP CTRL CONN request (disconnecting)");
-				if (gre_fd != -1 || pty_fd != -1)
-					syslog(LOG_WARNING, "CTRL: Request to close control connection when call is open, closing");
-				send_pptp_packet(clientSocket, rply_packet, rply_size);
-				goto leave_drop_call;
+      case -1:
+        send_packet = FALSE;
+        break;
 
-			case CALL_CLR_RQST:
-				if(pptpctrl_debug)
-					syslog(LOG_DEBUG, "CTRL: Received CALL CLR request (closing call)");
-				if (gre_fd == -1 || pty_fd == -1)
-					syslog(LOG_WARNING, "CTRL: Request to close call but call not open");
-				if (gre_fd != -1) {
-					FD_CLR(gre_fd, &fds);
-					close(gre_fd);
-				}
-				if (pty_fd != -1) {
-					FD_CLR(pty_fd, &fds);
-					close(pty_fd);
-				}
-				NOTE_VALUE(PAC, call_id_pair, htons(-1));
-				NOTE_VALUE(PNS, call_id_pair, htons(-1));
-				break;
+      case STOP_CTRL_CONN_RQST:
+        if (pptpctrl_debug)
+          syslog(LOG_DEBUG, "CTRL: Received STOP CTRL CONN request (disconnecting)");
+        if (gre)
+          syslog(LOG_WARNING, "CTRL: closing...");
+        send_pptp_packet(clientSocket, rply_packet, rply_size);
+        goto leave_drop_call;
 
-			case OUT_CALL_RQST:
-				/* for killing off the link later (ugly) */
-				NOTE_VALUE(PAC, call_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id);
-				NOTE_VALUE(PNS, call_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id_peer);
-				if (gre_fd != -1 || pty_fd != -1) {
-					syslog(LOG_WARNING, "CTRL: Request to open call when call is already open, closing");
-					if (gre_fd != -1) {
-						FD_CLR(gre_fd, &fds);
-						close(gre_fd);
-					}
-					if (pty_fd != -1) {
-						FD_CLR(pty_fd, &fds);
-						close(pty_fd);
-					}
-				}
-				/* Start the call */
-				syslog(LOG_INFO, "CTRL: Starting call (launching pppd, opening GRE)");
-				if ((pty_fd = startCall(pppaddrs)) > maxfd)
-					maxfd = pty_fd;
-				if ((gre_fd = pptp_gre_init(call_id_pair, pty_fd, inetaddrs)) > maxfd)
-					maxfd = gre_fd;
-				break;
+      case CALL_CLR_RQST:
+        if(pptpctrl_debug)
+          syslog(LOG_DEBUG, "CTRL: Received CALL CLR request (closing call)");
+        if (!gre || gre->gre_fd == -1 || gre->pty_fd == -1)
+          syslog(LOG_WARNING, "CTRL: Request to close call but call not open");
+        if (gre) {
+          if (gre->gre_fd != -1) {
+            close(gre->gre_fd);
+            gre->gre_fd = -1;
+          }
+          if (gre->pty_fd != -1) {
+            close(gre->pty_fd);
+            gre->pty_fd = -1;
+          }
+          NOTE_VALUE(PAC, call_id_pair, htons(-1));
+          NOTE_VALUE(PNS, call_id_pair, htons(-1));
+          break;
+        }
 
-			case ECHO_RPLY:
-				if (echo_wait == TRUE && ((struct pptp_echo_rply *) (packet))->identifier == echo_count)
-					echo_wait = FALSE;
-				else
-					syslog(LOG_WARNING, "CTRL: Unexpected ECHO REPLY packet");
-				/* FALLTHRU */
-			case SET_LINK_INFO:
-				send_packet = FALSE;
-				break;
+      case OUT_CALL_RQST:
+        /* for killing off the link later (ugly) */
+        NOTE_VALUE (PAC, call_id_pair,
+                    ((struct pptp_out_call_rply *) (rply_packet))->call_id);
+        NOTE_VALUE(PNS, call_id_pair, ((struct pptp_out_call_rply *)
+                                       (rply_packet))->call_id_peer);
+        max_window_size =
+          ntoh16 (((struct pptp_out_call_rqst *)packet)->pckt_recv_size);
+        if (gre) {
+          syslog(LOG_WARNING, "CTRL: Request to open call when call is already open, closing");
+          if (gre->gre_fd != -1) {
+            close (gre->gre_fd);
+            gre->gre_fd = -1;
+          }
+          if (gre->pty_fd != -1) {
+            close (gre->pty_fd);
+            gre->pty_fd = -1;
+          }
+          free (gre);
+        }
+        /* Start the call */
+        syslog(LOG_INFO, "CTRL: Starting call (launching pppd, opening GRE)");
+        if ((new_pty_fd = startCall(pppaddrs)) > maxfd)
+          maxfd = new_pty_fd;
+        gre = pptp_gre_init(call_id_pair, max_window_size,
+                            new_pty_fd, inetaddrs);
+        if (!gre) {
+          syslog (LOG_ERR, "pptp_gre_init failed; exiting");
+          exit (-1);
+        }
 
-			case CALL_DISCONN_NTFY:
-			case STOP_CTRL_CONN_RPLY:
-				/* These don't generate replies.  Also they come from things we don't send in this section. */
-				syslog(LOG_WARNING, "CTRL: Got a reply to a packet we didn't send");
-				send_packet = FALSE;
-				break;
+        if (gre->gre_fd > maxfd)
+          maxfd = gre->gre_fd;
+        break;
 
-			/* Otherwise, the already-formed reply will do fine, so send it */
-			}
-		}
+      case ECHO_RPLY:
+        if (echo_wait == TRUE && ((struct pptp_echo_rply *) (packet))->identifier == echo_count)
+          echo_wait = FALSE;
+        else
+          syslog(LOG_WARNING, "CTRL: Unexpected ECHO REPLY packet");
+        /* FALLTHRU */
+      case SET_LINK_INFO:
+        send_packet = FALSE;
+        break;
 
-		/* send reply packet - this may block, but it should be very rare */
-		if (send_packet == TRUE && send_pptp_packet(clientSocket, rply_packet, rply_size) < 0) {
-			syslog(LOG_ERR, "CTRL: Error sending GRE, aborting call");
-			goto leave_clear_call;
-		}
+      case CALL_DISCONN_NTFY:
+      case STOP_CTRL_CONN_RPLY:
+        /* These don't generate replies.  Also they come from things
+           we don't send in this section. */
+        syslog(LOG_WARNING, "CTRL: Got a reply to a packet we didn't send");
+        send_packet = FALSE;
+        break;
 
-		/* waiting for echo reply and curtime - echo_time > max wait */
-		if (echo_wait == TRUE && (time(NULL) - echo_time) > MAX_ECHO_WAIT) {
-			syslog(LOG_INFO, "CTRL: Session timed out, ending call");
-			goto leave_clear_call;
-		}
-	}
-	/* Finished! :-) */
+        /* Otherwise, the already-formed reply will do fine, so send it */
+      }
+    }
+
+    /* send reply packet - this may block, but it should be very rare */
+    if (send_packet == TRUE
+        && send_pptp_packet(clientSocket, rply_packet, rply_size) < 0) {
+      syslog(LOG_ERR, "CTRL: Error sending GRE, aborting call");
+      goto leave_clear_call;
+    }
+
+    /* waiting for echo reply and curtime - echo_time > max wait */
+    if (echo_wait == TRUE && (time(NULL) - echo_time) > MAX_ECHO_WAIT) {
+      syslog(LOG_INFO, "CTRL: Session timed out, ending call");
+      goto leave_clear_call;
+    }
+  }
+  /* Finished! :-) */
 leave_drop_call:
-	NOTE_VALUE(PAC, call_id_pair, htons(-1));
-	NOTE_VALUE(PNS, call_id_pair, htons(-1));
-	close(clientSocket);
+  NOTE_VALUE(PAC, call_id_pair, htons(-1));
+  NOTE_VALUE(PNS, call_id_pair, htons(-1));
+  close(clientSocket);
 leave_clear_call:
-	/* leave clientSocket and call_id_pair alone for bail() */
-	if (gre_fd != -1)
-		close(gre_fd);
-	if (pty_fd != -1)
-		close(pty_fd);
-	return;
+  /* leave clientSocket and call_id_pair alone for bail() */
+  if (gre->gre_fd != -1)
+    close(gre->gre_fd);
+  if (gre->pty_fd != -1)
+    close(gre->pty_fd);
+  return;
 #ifdef init
 #undef init
 #endif
@@ -452,9 +482,17 @@ static void bail(int sigraised)
 		tv.tv_usec = 0;
 
 		/* Wait for STOP CTRL CONN RQST or RPLY */
-		while (select(clientSocket + 1, &connSet, NULL, NULL, &tv) == 1) {
-			switch((pkt = read_pptp_packet(clientSocket, packet, rply_packet, &rply_size))) {
-			case STOP_CTRL_CONN_RQST:
+		while (select(clientSocket + 1, &connSet, NULL, NULL,
+                              &tv) == 1) {
+                  pkt = read_pptp_packet(clientSocket, packet,
+                                         rply_packet, &rply_size);
+                  if (pkt == 0) {
+                    syslog (LOG_WARNING,
+                            "EOF reading control message");
+                    break;
+                  }
+                  else switch (pkt) {
+                        case STOP_CTRL_CONN_RQST:
 				send_pptp_packet(clientSocket, rply_packet, rply_size);
 				goto skip;
 			case CALL_CLR_RQST:
@@ -462,6 +500,12 @@ static void bail(int sigraised)
 				break;
 			case STOP_CTRL_CONN_RPLY:
 				goto skip;
+			case 0:
+				syslog(LOG_WARNING, "CTRL: Client already disconnected control stream");
+				goto skip;
+			case -1:
+				syslog(LOG_WARNING, "CTRL: Error reading control stream during disconnect, aborting");
+				goto skip;
 			default:
 				syslog(LOG_WARNING, "CTRL: Unexpected control message %d in disconnect sequence", pkt);
 			}
@@ -491,6 +535,7 @@ static int startCall(char **pppaddrs)
 {
 	/* PTY/TTY pair for talking to PPPd */
 	int pty_fd, tty_fd;
+        int flags;
 #if BSDUSER_PPP || SLIRP
 	int sockfd[2];
 
@@ -558,6 +603,13 @@ static int startCall(char **pppaddrs)
 		exit(3);
 	}
 	close(tty_fd);
+
+        flags = fcntl (pty_fd, F_GETFL, 0);
+        if (flags < 0)
+          syslog (LOG_ERR, "CTRL: Couldn't get flags for TTY!");
+        else if (fcntl (pty_fd, F_SETFL, flags | O_NONBLOCK) < 0)
+          syslog (LOG_ERR, "CTRL: Cound't set flags for TTY!");
+
 	return pty_fd;
 }
 
